from pathlib import Path
from typing import Dict, List, Type, Union, cast

from graphql import (
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLUnionType,
    build_schema,
    extend_schema,
    parse,
)

from .enum import register_enums
from .federation import (
    federation_entity_type_defs,
    federation_service_type_defs,
    get_entity_types,
    purge_schema_directives,
    remove_subscription,
    resolve_entities,
)
from .resolver import register_resolvers
from .scalar import register_scalars
from .schema_visitor import SchemaDirectiveVisitor
from .utils import join_type_defs


def make_schema(
    type_defs: Union[str, List[str]],
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
    no_location: bool = False,
    experimental_fragment_variables: bool = False,
    federation: bool = False,
    directives: Dict[str, Type[SchemaDirectiveVisitor]] = None,
) -> GraphQLSchema:
    if isinstance(type_defs, list):
        type_defs = join_type_defs(type_defs)

    if federation:
        # Remove custom schema directives (to avoid apollo-gateway crashes).
        sdl = purge_schema_directives(type_defs)

        # remove subscription because Apollo Federation not support subscription yet.
        # type_defs = remove_subscription(type_defs)

        type_defs = join_type_defs([type_defs, federation_service_type_defs])
        schema = build_schema(
            type_defs, assume_valid, assume_valid_sdl, no_location, experimental_fragment_variables
        )
        entity_types = get_entity_types(schema)
        if entity_types:
            schema = extend_schema(schema, parse(federation_entity_type_defs))

            # Add _entities query.
            entity_type = schema.get_type("_Entity")
            if entity_type:
                entity_type = cast(GraphQLUnionType, entity_type)
                entity_type.types = entity_types

            query_type = schema.get_type("Query")
            if query_type:
                query_type = cast(GraphQLObjectType, query_type)
                query_type.fields["_entities"].resolve = resolve_entities

        # Add _service query.
        query_type = schema.get_type("Query")
        if query_type:
            query_type = cast(GraphQLObjectType, query_type)
            query_type.fields["_service"].resolve = lambda _service, info: {"sdl": sdl}
    else:
        schema = build_schema(
            type_defs, assume_valid, assume_valid_sdl, no_location, experimental_fragment_variables
        )

    register_resolvers(schema)
    register_enums(schema)
    register_scalars(schema)

    if directives:
        SchemaDirectiveVisitor.visit_schema_directives(schema, directives)
    return schema


def make_schema_from_file(
    file: str,
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
    no_location: bool = False,
    experimental_fragment_variables: bool = False,
    federation: bool = False,
    directives: Dict[str, Type[SchemaDirectiveVisitor]] = None,
) -> GraphQLSchema:
    with open(file, 'r') as f:
        schema = make_schema(
            f.read(),
            assume_valid,
            assume_valid_sdl,
            no_location,
            experimental_fragment_variables,
            federation,
            directives,
        )
        return schema


def parse_from_file(file: Path):
    with file.open('r') as f:
        type_defs = f.read()
        parse(type_defs)
        return type_defs


base_type_defs = """
type Query
type Mutation
"""


def make_schema_from_path(
    path: str,
    assume_valid: bool = False,
    assume_valid_sdl: bool = False,
    no_location: bool = False,
    experimental_fragment_variables: bool = False,
    federation: bool = False,
    directives: Dict[str, Type[SchemaDirectiveVisitor]] = None,
):
    p = Path(path)
    if p.is_file():
        type_defs = parse_from_file(p)
    elif p.is_dir():
        type_defs = [base_type_defs]
        for file in p.glob('*.graphql'):
            type_defs.extend([parse_from_file(file)])
    else:
        raise RuntimeError('path: expect a file or directory!')

    return make_schema(
        type_defs,
        assume_valid,
        assume_valid_sdl,
        no_location,
        experimental_fragment_variables,
        federation,
        directives,
    )


from typing import Any, Callable, Dict, List, Optional, Union

import graphql
from graphql.execution.execute import get_field_def
from graphql.execution.values import get_variable_values
from graphql.pyutils import AwaitableOrValue, FrozenList, Path, Undefined, inspect

from .middleware import MiddlewareManager


class ExecutionContext(graphql.ExecutionContext):
    middleware_manager: MiddlewareManager

    @classmethod
    def build(
        cls,
        schema: graphql.GraphQLSchema,
        document: graphql.DocumentNode,
        root_value: Any = None,
        context_value: Any = None,
        raw_variable_values: Optional[Dict[str, Any]] = None,
        operation_name: Optional[str] = None,
        field_resolver: Optional[graphql.GraphQLFieldResolver] = None,
        type_resolver: Optional[graphql.GraphQLTypeResolver] = None,
        middleware: Optional[graphql.Middleware] = None,
        is_awaitable: Optional[Callable[[Any], bool]] = None,
    ) -> Union[List[graphql.GraphQLError], "ExecutionContext"]:
        """Build an execution context

        Constructs a ExecutionContext object from the arguments passed to execute, which
        we will pass throughout the other execution methods.

        Throws a GraphQLError if a valid execution context cannot be created.

        For internal use only.
        """
        operation: Optional[graphql.OperationDefinitionNode] = None
        fragments: Dict[str, graphql.FragmentDefinitionNode] = {}
        middleware_manager: Optional[graphql.MiddlewareManager] = None
        if middleware is not None:
            if isinstance(middleware, (list, tuple)):
                middleware_manager = MiddlewareManager(*middleware)
            elif isinstance(middleware, MiddlewareManager):
                middleware_manager = middleware
            else:
                raise TypeError(
                    "Middleware must be passed as a list or tuple of functions"
                    " or objects, or as a single MiddlewareManager object."
                    f" Got {inspect(middleware)} instead."
                )

        for definition in document.definitions:
            if isinstance(definition, graphql.OperationDefinitionNode):
                if operation_name is None:
                    if operation:
                        return [
                            graphql.GraphQLError(
                                "Must provide operation name"
                                " if query contains multiple operations."
                            )
                        ]
                    operation = definition
                elif definition.name and definition.name.value == operation_name:
                    operation = definition
            elif isinstance(definition, graphql.FragmentDefinitionNode):
                fragments[definition.name.value] = definition

        if not operation:
            if operation_name is not None:
                return [graphql.GraphQLError(f"Unknown operation named '{operation_name}'.")]
            return [graphql.GraphQLError("Must provide an operation.")]

        coerced_variable_values = get_variable_values(
            schema,
            operation.variable_definitions or FrozenList(),
            raw_variable_values or {},
            max_errors=50,
        )

        if isinstance(coerced_variable_values, list):
            return coerced_variable_values  # errors

        return cls(
            schema,
            fragments,
            root_value,
            context_value,
            operation,
            coerced_variable_values,  # coerced values
            field_resolver or graphql.default_field_resolver,
            type_resolver or graphql.default_type_resolver,
            [],
            middleware_manager,
            is_awaitable,
        )

    def resolve_field(
        self,
        parent_type: graphql.GraphQLObjectType,
        source: Any,
        field_nodes: List[graphql.FieldNode],
        path: Path,
    ) -> AwaitableOrValue[Any]:
        """Resolve the field on the given source object.

        In particular, this figures out the value that the field returns by calling its
        resolve function, then calls complete_value to await coroutine objects,
        serialize scalars, or execute the sub-selection-set for objects.
        """
        field_node = field_nodes[0]
        field_name = field_node.name.value

        field_def = get_field_def(self.schema, parent_type, field_name)
        if not field_def:
            return Undefined

        resolve_fn = field_def.resolve or self.field_resolver

        if self.middleware_manager:
            resolve_fn = self.middleware_manager.get_field_resolver_by_parent(
                resolve_fn, parent_type.name, field_name
            )

        info = self.build_resolve_info(field_def, field_nodes, parent_type, path)

        # Get the resolve function, regardless of if its result is normal or abrupt
        # (error).
        result = self.resolve_field_value_or_error(field_def, field_nodes, resolve_fn, source, info)

        return self.complete_value_catching_error(field_def.type, field_nodes, info, path, result)



from functools import partial, reduce
from inspect import isfunction
from typing import Callable, Iterator, Any, Dict, List

from graphql.execution.middleware import MiddlewareManager as BaseMiddlewareManager

GraphQLFieldResolver = Callable[..., Any]


class MiddlewareManager(BaseMiddlewareManager):
    middlewares: dict
    exclude: List[str]

    def __init__(self, middlewares: Dict[str, list], exclude: List[str] = None):
        assert isinstance(middlewares, dict), f'MiddlewareManager expected dict, not {type(middlewares)}'
        self.middlewares = {
            key: list(get_middleware_resolvers(value)) if value else None for key, value in middlewares.items()
        }
        self._cached_resolvers = {}
        self.exclude = exclude or []

    def get_field_resolver_by_parent(
        self, field_resolver: GraphQLFieldResolver, parent_type: str, field_name: str
    ) -> GraphQLFieldResolver:
        field = f'{parent_type}.{field_name}'
        if field in self.exclude or parent_type not in self.middlewares and field not in self.middlewares:
            return field_resolver

        if field_resolver not in self._cached_resolvers:
            middlewares = self.middlewares[parent_type] if parent_type in self.middlewares else self.middlewares[field]
            self._cached_resolvers[field_resolver] = reduce(
                lambda chained_fns, next_fn: partial(next_fn, chained_fns),
                middlewares,
                field_resolver,
            )
        return self._cached_resolvers[field_resolver]


def get_middleware_resolvers(middlewares: list) -> Iterator[Callable]:
    """Get a list of resolver functions from a list of classes or functions."""
    for middleware in middlewares:
        if isfunction(middleware):
            yield middleware
        else:  # middleware provided as object with 'resolve' method
            resolver_func = getattr(middleware, "resolve", None)
            if resolver_func is not None:
                yield resolver_func



from .enum import enum_type  # noqa
from .execute import ExecutionContext  # noqa
from .middleware import MiddlewareManager  # noqa
from .parser import parse_info, FieldMeta, parse_node  # noqa
from .resolver import (  # noqa
    field_resolver,
    mutate,
    query,
    reference_resolver,
    subscribe,
    type_resolver,
)
from .scalar import scalar_type  # noqa
from .schema import make_schema, make_schema_from_file, make_schema_from_path  # noqa
from .schema_visitor import SchemaDirectiveVisitor  # noqa
from .utils import gql  # noqa

__version__ = '0.2.4'
